//Copyright INRIA/CNRS
//
//Petr Popov and Sergei Grudinin, 2012
//sergei.grudinin@inria.fr
//
//This software is a computer program whose purpose is to compute
//the root mean sum of squared deviations (RMSDs) corresponding
//to a rigid-body motion in constatnt time.
//
//This software is governed by the CeCILL  license under French law and
//abiding by the rules of distribution of free software.  You can  use,
//modify and/ or redistribute the software under the terms of the CeCILL
//license as circulated by CEA, CNRS and INRIA at the following URL
//"http://www.cecill.info".
//
//As a counterpart to the access to the source code and  rights to copy,
//modify and redistribute granted by the license, users are provided only
//with a limited warranty  and the software's author,  the holder of the
//economic rights,  and the successive licensors  have only  limited
//liability.
//
//In this respect, the user's attention is drawn to the risks associated
//with loading,  using,  modifying and/or developing or reproducing the
//software by the user in light of its specific status of free software,
//that may mean  that it is complicated to manipulate,  and  that  also
//therefore means  that it is reserved for developers  and  experienced
//professionals having in-depth computer knowledge. Users are therefore
//encouraged to load and test the software's suitability as regards their
//requirements in conditions enabling the security of their systems and/or
//data to be ensured and,  more generally, to use and operate it in the
//same conditions as regards security.
//
//The fact that you are presently reading this means that you have had
//knowledge of the CeCILL license and that you accept its terms.
#pragma once
#include <string>
#include <cstring>
#include <map>
#include <iostream>
#include <stdio.h>

typedef double vector3[3]; 

class cPDBAtom {

	public:

		cPDBAtom() {};
		cPDBAtom(std::string &pdbLine);
		~cPDBAtom() {};

		void initialize (const std::string &pdbLine);

		// weiths of most common atoms in pdb files
		float assignWeight (int symbol) {
			int upSym = toupper(symbol);
			if (upSym <'O') {
				if (upSym <'H') {
					if (upSym=='C') return 12.0107;	
				} else {
					if (upSym=='H') return 1.00794;  
					if (upSym=='N') return 14.0067;	
				}
			} else {
				if (upSym <'P') {
					if (upSym=='O') return 15.9994;
				} else {
					if (upSym=='P') return 30.973761; 
					if (upSym=='S') return 32.065;	
				}
			}
			return 1;  
		}

		char name[5];
		vector3 coords;
		double weight;
		
		friend std::ostream& operator<<(std::ostream &o, cPDBAtom *atom);
		friend std::istream& operator>>(std::istream & i, cPDBAtom *);

};

cPDBAtom::cPDBAtom (std::string &pdbLine) {

	// string should be : "ATOM   1601  N   PRO B 583      24.398  36.009 -31.776  1.00 39.22"
	try {

		sscanf(pdbLine.c_str(), "%*s%*d%s%*s%*s%*d%lf%lf%lf", name, coords, coords+1, coords+2 );
		weight = assignWeight(name[0]);

	}
	catch (...) {
		std::cerr<<"Incorrect input of coordinates..."<<std::endl;
	}
}

void cPDBAtom::initialize (const std::string &pdbLine) {

	// string should be : "ATOM   1601  N   PRO B 583      24.398  36.009 -31.776  1.00 39.22"
	try {

		sscanf(pdbLine.c_str(), "%*s%*d%s%*s%*s%*d%lf%lf%lf", name, coords, coords+1, coords+2 );
		weight = assignWeight(name[0]);

	}
	catch (...) {
		std::cerr<<"Incorrect input of coordinates..."<<std::endl;
	}
}

std::ostream &operator <<(std::ostream &o, cPDBAtom &atom) {
	return o<<atom.name<<' '
			<<' '<<atom.coords[0]<<' '<<atom.coords[1]<<' '<<atom.coords[2]
			<<' '<<atom.weight<<std::endl;
};