//Copyright INRIA/CNRS
//
//Neveu, Emilie; Popov, Petr; Hoffmann, Alexandre; Grudinin, Sergei. 2017
//sergei.grudinin@inria.fr
//
//This software is a computer program whose purpose is to compute
//the root mean sum of squared deviations (RMSDs) corresponding 
//to a rigid-body motion in constatnt time.
//
//This software is governed by the CeCILL-B  license (fully BSD-compatible)
//under French law and abiding by the rules of distribution of free software.
//You can  use, modify and/ or redistribute the software under the terms
//of the CeCILL-B license as circulated by CEA, CNRS and INRIA at the
//following URL "http://www.cecill.info".
//
//As a counterpart to the access to the source code and  rights to copy,
//modify and redistribute granted by the license, users are provided only
//with a limited warranty  and the software's author,  the holder of the
//economic rights,  and the successive licensors  have only  limited
//liability.
//
//In this respect, the user's attention is drawn to the risks associated
//with loading,  using,  modifying and/or developing or reproducing the
//software by the user in light of its specific status of free software,
//that may mean  that it is complicated to manipulate,  and  that  also
//therefore means  that it is reserved for developers  and  experienced
//professionals having in-depth computer knowledge. Users are therefore
//encouraged to load and test the software's suitability as regards their
//requirements in conditions enabling the security of their systems and/or
//data to be ensured and,  more generally, to use and operate it in the
//same conditions as regards security.
//
//The fact that you are presently reading this means that you have had
//knowledge of the CeCILL-B license and that you accept its terms.

/*
 *=================================================
 *==== example of getMSDTwoTransforms() function ====
 *=================================================
 */

#include "RapidRMSD.hpp" // main library
#include <stdlib.h>
#include <vector>
#include <iostream>
#include <math.h> // for sqrt
#include <time.h> // for srand

typedef double vector3[3]; 

void computeMatrix (double* m, double *q) {

	if (m && q) {
		double a = q[3];
		double b = q[0];
		double c = q[1];
		double d = q[2];

		m[0] = a*a+b*b-c*c-d*d; m[4] = a*a-b*b+c*c-d*d; m[8] = a*a-b*b-c*c+d*d;
		m[1] = 2*b*c-2*a*d; m[3] = 2*b*c+2*a*d;
		m[2] = 2*b*d+2*a*c; m[6] = 2*b*d-2*a*c;
		m[5] = 2*c*d-2*a*b; m[7] = 2*c*d+2*a*b;

		return;
	}
	else {
		std::cerr << "Error"<<std::endl;
		exit(0);
	}
}

double getLinearRMSD(vector3 *coords, double* m1, double *m2, vector3 T1, vector3 T2, int N, double *weights=NULL) {

	double Mass = 0 ;
	
	double msd = 0;
	double x,y,z;
	vector3 r1,r2;
	if (weights) {
		for (int i=0; i<N; ++i) {

			Mass+=weights[i];

			x=coords[i][0];
			y=coords[i][1];
			z=coords[i][2];

			r1[0] = m1[0]*x+m1[1]*y+m1[2]*z;
			r1[1] = m1[3]*x+m1[4]*y+m1[5]*z;
			r1[2] = m1[6]*x+m1[7]*y+m1[8]*z;

			r1[0]+=T1[0];
			r1[1]+=T1[1];
			r1[2]+=T1[2];

			r2[0] = m2[0]*x+m2[1]*y+m2[2]*z;
			r2[1] = m2[3]*x+m2[4]*y+m2[5]*z;
			r2[2] = m2[6]*x+m2[7]*y+m2[8]*z;

			r2[0]+=T2[0];
			r2[1]+=T2[1];
			r2[2]+=T2[2];

			msd+= weights[i]* ((r1[0]-r2[0])*(r1[0]-r2[0])+(r1[1]-r2[1])*(r1[1]-r2[1])+(r1[2]-r2[2])*(r1[2]-r2[2]));			
		}
		msd/=Mass;
	}
	else {
		for (int i=0; i<N; ++i) {

			x=coords[i][0];
			y=coords[i][1];
			z=coords[i][2];

			r1[0] = m1[0]*x+m1[1]*y+m1[2]*z;
			r1[1] = m1[3]*x+m1[4]*y+m1[5]*z;
			r1[2] = m1[6]*x+m1[7]*y+m1[8]*z;

			r1[0]+=T1[0];
			r1[1]+=T1[1];
			r1[2]+=T1[2];

			r2[0] = m2[0]*x+m2[1]*y+m2[2]*z;
			r2[1] = m2[3]*x+m2[4]*y+m2[5]*z;
			r2[2] = m2[6]*x+m2[7]*y+m2[8]*z;

			r2[0]+=T2[0];
			r2[1]+=T2[1];
			r2[2]+=T2[2];

			msd+= ((r1[0]-r2[0])*(r1[0]-r2[0])+(r1[1]-r2[1])*(r1[1]-r2[1])+(r1[2]-r2[2])*(r1[2]-r2[2]));			
		}
		msd/=N;
	}
	return msd;
}


int main() {

	std::cout<<"example of getMSDTwoTransforms() function\n";

	srand ( time(NULL) );

	//test coordinates
	int N = 20;
	vector3 *coords = new vector3[N];
		
	coords[0][0] = 1.941; coords[0][1] = -28.006; coords[0][2] =  27.760  ;  
	coords[1][0] = 2.783; coords[1][1] = -28.083; coords[1][2] =  26.553  ;  
	coords[2][0] = 2.110; coords[2][1] = -27.700; coords[2][2] =  25.235  ;  
	coords[3][0] = 2.801; coords[3][1] = -27.237; coords[3][2] =  24.337  ;  
	coords[4][0] = 3.276; coords[4][1] = -29.495; coords[4][2] =  26.288  ;  
	coords[5][0] = 3.603; coords[5][1] = -30.188; coords[5][2] =  27.574  ;  
	coords[6][0] = 2.380; coords[6][1] = -30.852; coords[6][2] =  28.025  ;  
	coords[7][0] = 1.978; coords[7][1] = -31.935; coords[7][2] =  27.576  ;  
	coords[8][0] = 1.821; coords[8][1] = -30.061; coords[8][2] =  28.894  ;  
	coords[9][0] = 0.800; coords[9][1] = -28.015; coords[9][2] =  25.074  ;  
	coords[10][0] = 0.045; coords[10][1] = -27.649; coords[10][2] =  23.874  ;  
	coords[11][0] =-0.264; coords[11][1] = -26.184; coords[11][2] =  23.771  ;  
	coords[12][0] =-0.183; coords[12][1] = -25.585; coords[12][2] =  22.710  ; 
	coords[13][0] =-1.170; coords[13][1] = -28.520; coords[13][2] =  23.699  ;  
	coords[14][0] =-0.762; coords[14][1] = -29.955; coords[14][2] =  23.503  ; 
	coords[15][0] =-0.834; coords[15][1] = -30.953; coords[15][2] =  24.417  ;  
	coords[16][0] =-0.289; coords[16][1] = -30.563; coords[16][2] =  22.315  ; 
	coords[17][0] =-0.413; coords[17][1] = -32.140; coords[17][2] =  23.903  ;  
	coords[18][0] =-0.203; coords[18][1] = -31.945; coords[18][2] =  22.564  ;  
	coords[19][0] =-0.126; coords[19][1] = -30.093; coords[19][2] =  21.056  ;  

	RMSD<vector3> rmsd;
	rmsd.init(coords, N); 	// RMSD init

	double qLength;
	double q[4];
	double m1[9], m2[9];
	vector3 T1, T2;
	
	for (int i=0; i<5; ++i) {
		
		T1[0] = (rand()%5);
		T1[1] = (rand()%5);
		T1[2] = (rand()%5);

		T2[0] = (rand()%5);
		T2[1] = (rand()%5);
		T2[2] = (rand()%5);
		
		q[0]= (rand()%100);
		q[1]= (rand()%100);
		q[2]= (rand()%100);
		q[3]= (rand()%100);
		q[0]/=100; q[1]/=100; q[2]/=100; q[3]/=100;
		//length of net quaternion
		qLength = sqrt(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]+q[3]*q[3]);
		//normalize it
		q[0]/=qLength;
		q[1]/=qLength;
		q[2]/=qLength;
		q[3]/=qLength;
		computeMatrix (m1, q);


		q[0]= (rand()%100);
		q[1]= (rand()%100);
		q[2]= (rand()%100);
		q[3]= (rand()%100);
		q[0]/=100; q[1]/=100; q[2]/=100; q[3]/=100;
		//length of net quaternion
		qLength = sqrt(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]+q[3]*q[3]);
		//normalize it
		q[0]/=qLength;
		q[1]/=qLength;
		q[2]/=qLength;
		q[3]/=qLength;
		computeMatrix (m2, q);

		std::cout<<"fast rmsd: "<<sqrt(rmsd.getMSDTwoTransforms(m1, m2, T1, T2)); // RMSD function call
		std::cout<<" linear RMSD: "<<sqrt(getLinearRMSD(coords, m1, m2, T1, T2, N))<<std::endl; // comparison with a direct evaluation
	}
		
	delete [] coords;

}
