//Copyright INRIA/CNRS
//
//Neveu, Emilie; Popov, Petr; Hoffmann, Alexandre; Grudinin, Sergei. 2017
//sergei.grudinin@inria.fr
//
//This software is a computer program whose purpose is to compute
//the root mean sum of squared deviations (RMSDs) corresponding 
//to a rigid-body motion in constatnt time.
//
//This software is governed by the CeCILL-B  license (fully BSD-compatible)
//under French law and abiding by the rules of distribution of free software.
//You can  use, modify and/ or redistribute the software under the terms
//of the CeCILL-B license as circulated by CEA, CNRS and INRIA at the
//following URL "http://www.cecill.info".
//
//As a counterpart to the access to the source code and  rights to copy,
//modify and redistribute granted by the license, users are provided only
//with a limited warranty  and the software's author,  the holder of the
//economic rights,  and the successive licensors  have only  limited
//liability.
//
//In this respect, the user's attention is drawn to the risks associated
//with loading,  using,  modifying and/or developing or reproducing the
//software by the user in light of its specific status of free software,
//that may mean  that it is complicated to manipulate,  and  that  also
//therefore means  that it is reserved for developers  and  experienced
//professionals having in-depth computer knowledge. Users are therefore
//encouraged to load and test the software's suitability as regards their
//requirements in conditions enabling the security of their systems and/or
//data to be ensured and,  more generally, to use and operate it in the
//same conditions as regards security.
//
//The fact that you are presently reading this means that you have had
//knowledge of the CeCILL-B license and that you accept its terms.

/*
 *================================================
 *==== example of getMSDQuaternion() function ====
 *================================================
 */

#include "RapidRMSD.hpp" // main library
#include <stdlib.h>
#include <vector>
#include "cPDBAtom.hpp" // for the PDB input
#include <math.h> // for sqrt
#include <time.h> // for srand

typedef double vector3[3]; 

double getLinearRMSDTwoMolecules(vector3 *coords, vector3 *coords1, double* q, vector3 T, int N, double *weights=NULL) {

	double Mass = 0 ;
	double A = q[3]*q[3]-(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]);
	
	double r = 0;
	double x,y,z;
	double x1,y1,z1;
	vector3 rtmp;
	if (weights) {
		for (int i=0; i<N; ++i) {

			Mass+=weights[i];

			x=coords[i][0];
			y=coords[i][1];
			z=coords[i][2];
			x1=coords1[i][0];
			y1=coords1[i][1];
			z1=coords1[i][2];

			double B = (x*q[0]+y*q[1]+z*q[2]);
			double cx = q[1]*z-q[2]*y;
			double cy = q[2]*x-q[0]*z;
			double cz = q[0]*y-q[1]*x;

			rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
			rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
			rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];

			rtmp[0]+=T[0];
			rtmp[1]+=T[1];
			rtmp[2]+=T[2];

			r+= weights[i]* ((x1-rtmp[0])*(x1-rtmp[0])+(y1-rtmp[1])*(y1-rtmp[1])+(z1-rtmp[2])*(z1-rtmp[2]));
		}
		r/=Mass;
	}
	else {
		for (int i=0; i<N; ++i) {

			x=coords[i][0];
			y=coords[i][1];
			z=coords[i][2];
			x1=coords1[i][0];
			y1=coords1[i][1];
			z1=coords1[i][2];

			double B = (x*q[0]+y*q[1]+z*q[2]);
			double cx = q[1]*z-q[2]*y;
			double cy = q[2]*x-q[0]*z;
			double cz = q[0]*y-q[1]*x;

			rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
			rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
			rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];

			rtmp[0]+=T[0];
			rtmp[1]+=T[1];
			rtmp[2]+=T[2];

			r+= ((x1-rtmp[0])*(x1-rtmp[0])+(y1-rtmp[1])*(y1-rtmp[1])+(z1-rtmp[2])*(z1-rtmp[2]));
		}
		r/=N;
	}
	return r;
}

int main() {

	srand( time(NULL) );

	std::cout<<"example of getMSDQuaternion() function\n";

	//test coordinates
	//initialization with pdbLines
	std::vector<cPDBAtom> atomsBound, atomsUnbound;
	atomsBound.resize(20);
	atomsUnbound.resize(20);
	
	atomsBound[0].initialize("ATOM      4  N   GLU A   1       1.941  -2.006   3.760  1.00 25.22           N");
	atomsBound[1].initialize("ATOM      5  CA  GLU A   1       2.783  -2.083   2.553  1.00 29.88           C");
	atomsBound[2].initialize("ATOM      6  C   GLU A   1       2.110  -2.700   2.235  1.00 26.79           C");
	atomsBound[3].initialize("ATOM      7  O   GLU A   1       2.801  -2.237   1.337  1.00 22.16           O");
	atomsBound[4].initialize("ATOM      8  CB  GLU A   1       3.276  -2.495   2.288  1.00 25.90           C");
	atomsBound[5].initialize("ATOM      9  CG  GLU A   1       3.603  -0.188   3.574  1.00 48.92           C");
	atomsBound[6].initialize("ATOM     10  CD  GLU A   1       2.380  -0.852   4.025  1.00 29.36           C");
	atomsBound[7].initialize("ATOM     11  OE1 GLU A   1       1.978   1.935   3.576  1.00 28.67           O");
	atomsBound[8].initialize("ATOM     12  OE2 GLU A   1       1.821  -0.061   4.894  1.00 37.04           O");
	atomsBound[9].initialize("ATOM     13  N   TRP A   2       0.800  -2.015   2.074  1.00 20.72           N");
	atomsBound[10].initialize("ATOM     14  CA  TRP A   2       0.045  -3.649  -1.874  1.00 16.62           C");
	atomsBound[11].initialize("ATOM     15  C   TRP A   2      -0.264  -4.184  -1.771  1.00 18.21           C");
	atomsBound[12].initialize("ATOM     16  O   TRP A   2      -0.183  -5.585  -2.710  1.00 31.39           O");
	atomsBound[13].initialize("ATOM     17  CB  TRP A   2      -1.170  -2.520  -1.699  1.00 21.45           C");
	atomsBound[14].initialize("ATOM     18  CG  TRP A   2      -0.762  -1.955  -1.503  1.00 23.04           C");
	atomsBound[15].initialize("ATOM     19  CD1 TRP A   2      -0.834  -0.953   0.417  1.00 30.95           C");
	atomsBound[16].initialize("ATOM     20  CD2 TRP A   2      -0.289  -0.563  -2.315  1.00 18.69           C");
	atomsBound[17].initialize("ATOM     21  NE1 TRP A   2      -0.413  -2.140  -2.903  1.00 24.73           N");
	atomsBound[18].initialize("ATOM     22  CE2 TRP A   2      -0.203  -1.945  -3.564  1.00 27.52           C");
	atomsBound[19].initialize("ATOM     23  CE3 TRP A   2      -0.126  -0.093  -3.056  1.00 20.75           C");
	
	std::cout<<atomsBound.size()<<" atoms created"<<std::endl;
	
	//randomize displacements
	double maxShift = 5;
	for (unsigned int i=0; i<atomsBound.size(); i++) {
		atomsUnbound[i].coords[0] = atomsBound[i].coords[0] + (double(rand())/RAND_MAX - 0.5)*2*maxShift;
		atomsUnbound[i].coords[1] = atomsBound[i].coords[1] + (double(rand())/RAND_MAX - 0.5)*2*maxShift;
		atomsUnbound[i].coords[2] = atomsBound[i].coords[2] + (double(rand())/RAND_MAX - 0.5)*2*maxShift;
		atomsUnbound[i].weight = atomsBound[i].weight;
	}
	
	// compute coordinates and weights
	vector3 *coordsBound = new vector3[atomsBound.size()];
	vector3 *coordsUnbound = new vector3[atomsUnbound.size()];
	double *weightsBound = new double[atomsBound.size()];
	double *weightsUnbound = new double[atomsUnbound.size()];
	
	for (unsigned int i=0; i<atomsBound.size(); i++) {
		coordsBound[i][0] = atomsBound[i].coords[0];
		coordsBound[i][1] = atomsBound[i].coords[1];
		coordsBound[i][2] = atomsBound[i].coords[2];
		weightsBound[i] = atomsBound[i].weight;
	}
	
	for (unsigned int i=0; i<atomsUnbound.size(); i++) {
		coordsUnbound[i][0] = atomsUnbound[i].coords[0];
		coordsUnbound[i][1] = atomsUnbound[i].coords[1];
		coordsUnbound[i][2] = atomsUnbound[i].coords[2];
		weightsUnbound[i] = atomsUnbound[i].weight;
	}
	
	RMSD<vector3> rmsd;
	rmsd.init(coordsBound, coordsUnbound, atomsBound.size(), NULL); 	// RMSD init

	
	double qLength;
	double q[4];
	vector3 T;
	
	for (int i=0; i<5; ++i) {
		T[0] = double(rand())/RAND_MAX-0.5;
		T[1] = double(rand())/RAND_MAX-0.5;
		T[2] = double(rand())/RAND_MAX-0.5;
		q[0]= double(rand())/RAND_MAX;
		q[1]= double(rand())/RAND_MAX;
		q[2]= double(rand())/RAND_MAX;
		q[3]= double(rand())/RAND_MAX;
		q[0]/=100; q[1]/=100; q[2]/=100; q[3]/=100;
		//length of net quaternion
		qLength = sqrt(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]+q[3]*q[3]);
		//normalize it
		q[0]/=qLength;
		q[1]/=qLength;
		q[2]/=qLength;
		q[3]/=qLength;

		std::cout<<"fast rmsd: "<<sqrt(rmsd.getMSDQuaternionTwoMolecules(q, T)); // RMSD function call
		std::cout<<" linear RMSD: "<<sqrt(getLinearRMSDTwoMolecules(coordsBound, coordsUnbound, q, T, atomsBound.size(), NULL))<<std::endl; // comparison with a direct evaluation
	}
		
	delete [] coordsBound;
	delete [] coordsUnbound;
	delete [] weightsBound;
	delete [] weightsUnbound;
}


